﻿/*
 * ##############################################
 * CalcFunctions
 * Main part of the QuickCalculator
 * ##############################################
 * 
 * You may copy and distribute this code and the 
 * whole program as you want.
 * This code is not meant to be sold.
 * 
 * Written by Philipp Erler
 * Contact me for feedback, new version or other causes:
 * ph.erler@gmx.net
 * 
 * Version info:
 * Last change: 2012 | 04 | 29
 * .net-Framework: V. 4.0
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CalculationE
{
    static class CalcOperators
    {
        //operator enum
        public enum opIndex                     { plus, minus, mult, div, pow, radic, log, mod, rnd,
                                                  lshift, rshift, lrot, rrot, greater, lesser, grEqu, leEqu, 
                                                  equal, notEqu, bAnd, xor, bOr, lAnd, lOr};
        //operator strings for comparison
        static public string[] opCompareStr =   { "+", "-", "*", "/", "^", "root", "log", "mod", "rnd", 
                                                  "<<", ">>", "lrot", "rrot", ">", "<", ">=", "<=",
                                                  "==", "!=", "&", "xor", "|", "&&", "||" };
        //operator order values, the smaller the faster executed
        static public int[] opOrderVal =        { 9, 9, 10, 10, 11, 11, 11, 10, 12, 
                                                  8, 8, 8, 8, 7, 7, 7, 7,
                                                  6, 6, 5, 4, 3, 2, 1 };

        public struct operatorInfo                                     //for get next operator function
        {
            public int pos;
            public int length;
            public opIndex type;
        }

        #region Operators

        //give string, return next Operator
        static operatorInfo Get_Next_Op(string inputStr)
        {
            operatorInfo nextOp = new operatorInfo();

            //init
            nextOp.pos = -1;

            for (int i = 0; i <= opCompareStr.GetUpperBound(0); ++i)
            {
                int thisOpPos = -1;
                thisOpPos = inputStr.IndexOf(opCompareStr[i]);
                if (thisOpPos > -1)
                {
                    if (thisOpPos < nextOp.pos || nextOp.pos == -1)
                    {
                        nextOp.pos = thisOpPos;
                        nextOp.length = opCompareStr[i].Length;
                        nextOp.type = (opIndex)i;
                    }
                }
            }

            return nextOp;
        }

        //give string, return last Operator
        static operatorInfo Get_Last_Op(string inputStr)
        {
            operatorInfo lastOp = new operatorInfo();

            for (int i = 0; i <= opCompareStr.GetUpperBound(0); ++i)
            {
                int thisOpPos = -1;
                thisOpPos = inputStr.LastIndexOf(opCompareStr[i]);
                if (thisOpPos > -1)
                {
                    if (thisOpPos > lastOp.pos)
                    {
                        lastOp.pos = thisOpPos;
                        lastOp.length = opCompareStr[i].Length;
                        lastOp.type = (opIndex)i;
                    }
                }
            }

            return lastOp;
        }

        /// <summary>
        /// Give operator, return order value. Return -1 if invalid index. 
        /// </summary>
        /// <param name="operatorInput"></param>
        /// <returns></returns>
        static public int Get_Operator_Order_Value(CalculationElement operatorInput)
        {
            int normalOpVal = -1;

            //TODO: remove
            System.Diagnostics.Debug.Assert(operatorInput.NestedLevel > -1);

            if (operatorInput.Index < 0 || operatorInput.Index >= opOrderVal.Count())
                return -1;
            else
            {
                normalOpVal = opOrderVal[operatorInput.Index];
                //due to recursive algorithm considering the nestedlevel is not needed at the moment
                return normalOpVal + (opOrderVal.Count() + 1) * operatorInput.NestedLevel;
            }
        }

        /// <summary>
        /// Return Index of Operator. Return -1 if not found.
        /// </summary>
        /// <param name="assumeOp"></param>
        /// <returns></returns>
        static public int Get_Operator_Index(string assumeOp)
        {
            assumeOp = assumeOp.Trim().ToLower();

            for (int i = 0; i <= opCompareStr.GetUpperBound(0); ++i)
            {
                if (assumeOp == opCompareStr[i])
                    return i;
            }

            return -1;
        }

        #endregion

        //give 2 values and 1 operator, return calculation result
        static public CalculationElement Simple_Calculation(CalculationElement ValA, 
            CalculationElement Operator, CalculationElement ValB)
        {
            double calcResult;

            //interprete operands
            if (ValA.Type == CalculationElement.ElementType.Variable)
                ValA = ValA.Interprete_Variable();
            if (ValB.Type == CalculationElement.ElementType.Variable)
                ValB = ValB.Interprete_Variable();

            if (ValA.Type != CalculationElement.ElementType.Number || ValB.Type != CalculationElement.ElementType.Number ||
                Operator.Type != CalculationElement.ElementType.Operator)
                throw new Exception("Invalid operation");

            try
            {
                if (ValB.GetValue() == 0.0 && Operator.Index == (int)opIndex.div)
                    throw new Exception("Division by zero");

                if (Operator.Index == (int)opIndex.plus)            // +
                    calcResult = ValA.GetValue() + ValB.GetValue();
                else if (Operator.Index == (int)opIndex.minus)      // -
                    calcResult = ValA.GetValue() - ValB.GetValue();
                else if (Operator.Index == (int)opIndex.mult)       // *
                    calcResult = ValA.GetValue() * ValB.GetValue();
                else if (Operator.Index == (int)opIndex.div)        // /
                    calcResult = ValA.GetValue() / ValB.GetValue();
                else if (Operator.Index == (int)opIndex.pow)        // ^
                {
                    if (ValA.GetValue() < 0 && Math.Abs(ValB.GetValue()) < 1)
                        throw new Exception("Complex result of " + ValA.ToString() + " ^ " + ValB.ToString() + " = " + System.Math.Pow(Math.Abs(ValA.GetValue()), ValB.GetValue()).ToString() + " i");
                    else
                        calcResult = System.Math.Pow(ValA.GetValue(), ValB.GetValue());
                }
                else if (Operator.Index == (int)opIndex.radic)      // root
                {
                    if (ValA.GetValue() == 0)
                        throw new Exception("Division by zero");
                    else if (ValB.GetValue() < 0)
                        throw new Exception("Complex result of " + ValA.ToString() + " root " + ValB.ToString() + " = " + System.Math.Pow(Math.Abs(ValB.GetValue()), 1 / ValA.GetValue()).ToString() + " i");
                    else
                        calcResult = System.Math.Pow(ValB.GetValue(), (1 / ValA.GetValue()));
                }
                else if (Operator.Index == (int)opIndex.log)        // log
                    calcResult = System.Math.Log(ValB.GetValue(), ValA.GetValue());
                else if (Operator.Index == (int)opIndex.mod)        // mod
                    calcResult = ValA.GetValue() % ValB.GetValue();
                else if (Operator.Index == (int)opIndex.rnd)        // rnd
                    calcResult = Math.Round(ValB.GetValue(), System.Convert.ToInt32(Math.Round(ValA.GetValue())));
                else if (Operator.Index == (int)opIndex.lshift)     // lshift
                    calcResult = ValA.ToInt() << ValB.ToInt();
                else if (Operator.Index == (int)opIndex.rshift)     // rshift
                    calcResult = ValA.ToInt() >> ValB.ToInt();
                else if (Operator.Index == (int)opIndex.lrot)       // lrot
                {
                    int numShifts = ValB.ToInt();
                    int res = ValA.ToInt();
                    for (int i = 0; i < numShifts; ++i)
                    {
                        int firstBit = res & -2147483648;           // -2147483648 -> complement on two: 0b1000...
                        res = res << 1;
                        if (firstBit != 0)
                            ++res;                                  // set LSB
                    }
                    calcResult = res;
                }
                else if (Operator.Index == (int)opIndex.rrot)       // rrot
                {
                    int numShifts = ValB.ToInt();
                    int res = ValA.ToInt();
                    for (int i = 0; i < numShifts; ++i)
                    {
                        int lastBit = res & 1;                      // 1 -> 0b000...0001
                        res = res >> 1;                             // moving sign bit
                        if (res < 0)
                            res &= 2147483647;                      // remove sign bit
                        if (lastBit != 0)
                            res |= -2147483648;                     // set MSB
                    }
                    calcResult = res;
                }
                else if (Operator.Index == (int)opIndex.greater)    // greater
                    calcResult = System.Convert.ToDouble(ValA.GetValue() > ValB.GetValue());
                else if (Operator.Index == (int)opIndex.lesser)     // lesser
                    calcResult = System.Convert.ToDouble(ValA.GetValue() < ValB.GetValue());
                else if (Operator.Index == (int)opIndex.grEqu)      // grEqu
                    calcResult = System.Convert.ToDouble(ValA.GetValue() >= ValB.GetValue());
                else if (Operator.Index == (int)opIndex.leEqu)      // leEqu
                    calcResult = System.Convert.ToDouble(ValA.GetValue() <= ValB.GetValue());
                else if (Operator.Index == (int)opIndex.equal)      // equal
                    calcResult = System.Convert.ToDouble(ValA.GetValue() == ValB.GetValue());
                else if (Operator.Index == (int)opIndex.notEqu)     // not equal
                    calcResult = System.Convert.ToDouble(ValA.GetValue() != ValB.GetValue());
                else if (Operator.Index == (int)opIndex.bAnd)       // bAnd
                    calcResult = ValA.ToInt() & ValB.ToInt();
                else if (Operator.Index == (int)opIndex.xor)        // xor
                    calcResult = ValA.ToInt() ^ ValB.ToInt();
                else if (Operator.Index == (int)opIndex.bOr)        // bOr
                    calcResult = ValA.ToInt() | ValB.ToInt();
                else if (Operator.Index == (int)opIndex.lAnd)       // lAnd
                    calcResult = System.Convert.ToDouble(ValA.ToBool() && ValB.ToBool());
                else if (Operator.Index == (int)opIndex.lOr)        // lOr
                    calcResult = System.Convert.ToDouble(ValA.ToBool() || ValB.ToBool());
                else                                                // unknown
                    throw new Exception("Unknown operator " + Operator.ToString());

                //same nestedlevel as operator
                return new CalculationElement(calcResult, Operator.NestedLevel);
            }
            catch (Exception e)
            {
                //something strange has happened
                //vala or valb is an unknown variable (variable name was not replaced by its value because the variable was not set)
                try
                {
                    System.Convert.ToDouble(ValA.ToString());
                }
                catch (Exception)
                {
                    throw new Exception("Unknown variable " + ValA);
                }
                try
                {
                    System.Convert.ToDouble(ValB.ToString());
                }
                catch (Exception)
                {
                    throw new Exception("Unknown variable " + ValB);
                }

                throw e;
            }
        }
    }
}
